#!/usr/bin/perl
#use daveperl;
use strict;

sub dprintf1 {}
sub dprintf2 {}
sub dprint1 {}
sub dprint2 {}

sub str_to_coord
{
	my $str = shift;
    	dprintf1 "coord line: '$str'\n";
	my ($lon, $lat) = split(",", $str, 2);
	my $ret;
	$ret->{lon} = $lon;
	$ret->{lat} = $lat;
	die "bad lat: '$lat' '$str'" if !$lat;
	die "bad lon: '$lon' '$str'" if !$lon;
	return $ret;
}

sub coord_avg
{
	my @coords = @_;
	my $ret;
	foreach my $coord (@coords) {
		$ret->{lat} += $coord->{lat};
		$ret->{lon} += $coord->{lon};
	}
	my $size = scalar(@coords);
	$ret->{lon} /= $size;
	$ret->{lat} /= $size;
	return $ret;
}

sub cat_file_into_array
{
	my $filename = shift;;
	my $nr_lines = shift;
	my @contents;
	my $line;
	open FILE, "< $filename";
	while ($line = <FILE>) {
		chomp $line;
		push @contents, $line;
		last if defined $nr_lines && --$nr_lines <= 0;
	}
	close FILE;
	return @contents;
}


my @americas = cat_file_into_array("america.php");

my $name = undef;

my $middles;
while (scalar @americas) {
	my $line = shift @americas;
	chomp $line;
	if ($line =~ /<name>/) {
		$name = $line;
		$name =~ s/<[\/]?[a-z]+>//ig;
		$name =~ s/\s*//;
	}
	if ($line =~ /<coordinates>/
	    && defined $name) {
		my @coords = ();
		push @coords, str_to_coord(shift(@americas));
		push @coords, str_to_coord(shift(@americas));
		push @coords, str_to_coord(shift(@americas));
		push @coords, str_to_coord(shift(@americas));
		$middles->{$name.".img"} = coord_avg(@coords);
		$name = undef;
    	}

}

my $parts="east west north south";

my @areas = cat_file_into_array("all-areas");
map {s/\s//g} @areas;
#putting it in a hash and back gets rid of dups
my %h;
#also detect files that might be missing
map {$h{$_.".img"} = 1 if -e $_.".img";} @areas;
@areas = keys %h;
printf "areas len: %d total size: %s\n", scalar(@areas), asizestr(@areas);
sub cstr
{
	my $coor = shift;
	return sprintf("(%f, %f)", $coor->{lon}, $coor->{lat});
}

sub asize
{
	my $total;
	foreach my $area (@_) {
		$total += -s $area;
	}
	return $total;
}

sub asizestr
{
	my $s = asize(@_);
	return sprintf "%5.2f MB", ($s * 1.0 / (1<<20));
}

sub cmpit
{
	my $a = shift;
	my $b = shift;
	my $ll = shift;
	return $a->{$ll} <=> $b->{$ll};
}

sub cmplat { return cmpit(@_, "lat");}
sub cmplon { return cmpit(@_, "lon");}

sub sortlat
{
	return cmplat($middles->{$a}, $middles->{$b});
}
sub sortlon
{
	return cmplon($middles->{$a}, $middles->{$b});
}

my $MB = (1<<20);
my $max_bucket_size = $ARGV[0] * $MB;
my $target_bucket_size = $max_bucket_size;
my $target_pieces = 1;

my $total = asize(@areas);
if ($total > $target_bucket_size) {
	use POSIX;
	my $pieces = 1.0 * $total / $max_bucket_size;
	$target_pieces = POSIX::ceil($pieces);
	$target_bucket_size = $total / $target_pieces;
	printf("need %f pieces each: %f\n",
		$pieces, $max_bucket_size>>20);
}

my @even_buckets = [ () ];
@areas = (sort sortlon (@areas));
# this will split up evenly more or less
foreach my $area (@areas) {
	if (asize(@{$even_buckets[-1]}) + asize($area) > $target_bucket_size 
	    && scalar(@even_buckets) < $target_pieces) {
		push @even_buckets, [ () ];
	}
	push @{$even_buckets[-1]}, $area;
	dprintf2 "%s avg: %s (total: %s)\n", $area,
		cstr($middles->{$area}),
		asizestr(@{$even_buckets[-1]});

}

for (my $i = 0; $i < scalar(@even_buckets); $i++) {
	my @bucket = @{$even_buckets[$i]};
	printf "bucket[%d] len: %d size: %s\n", $i,
		scalar(@bucket), asizestr(@bucket);
}

sub ahash
{
	my @in = @_;
	my %hash;
	foreach my $entry (@in) {
		$hash{$entry} = 1;
	}
	return \%hash;
}

sub stealhigh
{
	return steal(0, @_);
}
sub steallow{
	return steal(1, @_);
}

sub steal
{
	my $low = shift;

	my $all_buckets_ref = shift;
	my $target_buckets_ref = shift;
	my $index = shift;

	my @all_buckets = @{$all_buckets_ref};
	my @this_bucket = @{$target_buckets_ref};

	$index += 1 if ! $low;
	$index -= 1 if $low;
	return if $index >= scalar(@all_buckets);
	return if $index < 0;

	my @steal_bucket = @{$all_buckets[$index]};
	my $stolen = undef;
	my $hash = ahash(@this_bucket);
	while (scalar(@steal_bucket)) {
		if ($low) {
			$stolen = pop(@steal_bucket);
		} else {
			$stolen = shift(@steal_bucket);
		}
		#rintf("stole: $stolen\n");
		last if (!defined($hash->{$stolen}));
	}
	return $stolen;
}

my @result_buckets;
for (my $i = 0; $i < scalar(@even_buckets); $i++) {
	my @newbucket = @{$even_buckets[$i]};
	my @highneighbor;
	my @lowneighbor;
	while (1) {
		my $stolenhigh = stealhigh(\@even_buckets, \@newbucket, $i);
		if (defined $stolenhigh) {
			last if asize(@newbucket) + asize($stolenhigh) > $max_bucket_size;
			push @newbucket, $stolenhigh;
			dprint1 "stole high: $stolenhigh\n";
		}
		my $stolenlow = steallow(\@even_buckets, \@newbucket, $i);
		if (defined $stolenlow) {
			last if asize(@newbucket) + asize($stolenlow) > $max_bucket_size;
			unshift @newbucket, $stolenlow;
			dprint1 "stole low: $stolenlow\n";
		}
		if (! defined $stolenlow  && ! defined $stolenhigh) {
			die "unable to fill bucket $i";
			last;
		}
	}
	my $left = $middles->{$newbucket[0]}->{lon};
	my $right = $middles->{$newbucket[-1]}->{lon};
	printf "newbucket[%d] (%f:%f) len: %d size: %s\n", $i, $left, $right,
		scalar(@newbucket), asizestr(@newbucket);
	push @result_buckets, \@newbucket;
}

printf("total size: %s\n", asizestr(@areas));

sub mkmap
{
	my $name = shift;
	my $size = shift;
	my @imgs = @_;
	my @cmd = qw(java -Xmx6000m -jar /home/daveh/mkgmap-stuff/mkgmap/dist/mkgmap.jar   --gmapsupp --routable --net);
	push @cmd, "--area-name=\"USA $name\"";
	printf("making $name expected size: %s...", asizestr(@imgs));
	#system("echo", @cmd, @imgs);
	system(@cmd, @imgs);
	my $fsize = (1.0*(-s "gmapsupp.img") / (1.0*1<<20));
	printf("done (actual size: %4.2fMB)\n", $fsize);
	rename("gmapsupp.img", "$size-$name.gmapsupp.img");
}

my $size_mb = $ARGV[0];

foreach my $bucket_ref (@result_buckets) {
	my @bucket = @{$bucket_ref};
	my $left = sprintf "%5.2f", $middles->{$bucket[0]}->{lon};
	my $right = sprintf "%5.2f", $middles->{$bucket[-1]}->{lon};
	my $name = "lon_".$left."_to_".$right;
	my $sstr = $size_mb."MB";
	mkmap($name, $sstr, @bucket);
}

#mkmap("west", $size_mb."MB", @left_bucket);
#mkmap("east", $size_mb."MB", @right_bucket);
#mkmap("central", $size_mb."MB", @middle_bucket);
